/*
 * kexec: Linux boots Linux
 *
 * Copyright (C) 2003,2004  Eric Biederman (ebiederm@xmission.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

.data
	.equ	MSR_K6_EFER,   0xC0000080
	.equ	EFER_LME,      0x00000100
	.equ	X86_CR4_PAE,   0x00000020
	.equ	CR0_PG,        0x80000000

	.globl setup16_start, setup16_end, setup16_size, setup16_align
	.globl setup16_regs
setup16_start:
_reloc = .
	.balign 16
	.code64
	/* Compute where I am running at */
	movq	setup16_start(%rip), %rbx

	/* Relocate the code */
	addq	%rbx, lreloc1(%rip)

	/* Fixup my real mode segment */
	movl	%ebx, %eax
	shrl	$4, %eax
	movw	%ax, 2 + realptr(%rip)

	/* Fixup the gdt */
	movq	%rbx, %rax
	andl	$0x00ffffff, %eax
	shlq	$16,  %rax

	movq	%rbx, %rdx
	andl	$0xff000000, %edx
	shlq	$32, %rdx
	orq	%rdx, %rax

	orq	%rax, 0x20 + gdt(%rip)
	orq	%rax, 0x28 + gdt(%rip)

	/* Setup the classic BIOS interrupt table at 0x0 */
	lidt	idtptr(%rip)
	
	/* Provide us with 32bit and 16bit segments that we can use */
	lgdt	gdt(%rip)
		
	/* Switch to 32bit compatiblity mode */
	ljmp	*lm_exit_addr(%rip)
lm_exit:
	.code32

	/* Disable paging */
	movl	%cr0, %eax
	andl	$~CR0_PG, %eax
	movl	%eax, %cr0

	/* Disable long mode */
	movl	$MSR_K6_EFER, %ecx
	rdmsr
	andl	$~EFER_LME, %eax
	wrmsr

	/* Disable PAE */
	xorl	%eax, %eax
	movl	%eax, %cr4

	/* Note we don't disable the a20 line, (this shouldn't be required)
	 * The code to do it is in kexec_test and it is a real pain.
	 * I will worry about that when I need it.
	 */
	
	/* Load 16bit data segments, to ensure the segment limits are set */
	movl	$0x28, %eax
	movl	%eax, %ds
	movl	%eax, %es
	movl	%eax, %ss
	movl	%eax, %fs
	movl	%eax, %gs

	/* switch to 16bit mode */
	ljmp	$0x20, $2f - _reloc
2:
	.code16
	/* Disable Paging and protected mode */
	/* clear the PG & PE bits of CR0 */
	movl	%cr0,%eax
	andl	$~((1 << 31)|(1<<0)),%eax
	movl	%eax,%cr0

	/* make intersegment jmp to flush the processor pipeline
	 * and reload %cs:%eip (to clear upper 16 bits of %eip).
	 */
	ljmp	*(realptr - _reloc)
3:
	/* we are in real mode now
	 * set up the real mode segment registers : %ds, $ss, %es
	 */
	/* Setup the data segment */
	movw	%cs, %ax
	movw	%ax, %ds

	/* Load the registers */
	movl	eax - _reloc, %eax
	movl	ebx - _reloc, %ebx
	movl	ecx - _reloc, %ecx
	movl	edx - _reloc, %edx
	movl	esi - _reloc, %esi
	movl	edi - _reloc, %esi
	movl	esp - _reloc, %esp
	movl	ebp - _reloc, %ebp
	movw	es - _reloc, %es
	movw	ss - _reloc, %ss
	movw	fs - _reloc, %fs
	movw	gs - _reloc, %gs
	movw	ds - _reloc, %ds

	/* Jump to the kernel entrypoint */
	ljmp	%cs:*(realdest - _reloc)

setup16_regs:	
eax:	.long	0x00000000
ebx:	.long	0x00000000
ecx:	.long	0x00000000
edx:	.long	0x00000000
esi:	.long	0x00000000
edi:	.long	0x00000000
esp:	.long	0x00000000
ebp:	.long	0x00000000
ds:	.word	0x0000
es:	.word	0x0000
ss:	.word	0x0000
fs:	.word	0x0000
gs:	.word	0x0000
realdest:
ip:	.word	0x0000
cs:	.word	0x0000

	.balign 16
realptr:		
	.word	3b - _reloc
	.word	0x0000
	
idtptr:
	/* 256 entry idt at 0 */
	.word	0x400 - 1
	.quad	0
	.word	0

gdt:	/* 0x00 unusable segment 
	 * 0x08 unused
	 * so use them as the gdt ptr
	 */
	.word	gdt_end - gdt - 1
lreloc1:
	.quad	gdt - _reloc
	.word	0, 0, 0

	/* 0x10 4GB flat code segment */
	.word	0xFFFF, 0x0000, 0x9A00, 0x00CF
	/* 0x18 4GB flat data segment */
	.word	0xFFFF, 0x0000, 0x9200, 0x00CF

	/* 0x20 16 bit real mode code segment */
	.word	0xffff, 0x0000
	.byte	0x00, 0x9b, 0x00, 0x00

	/* 0x28 16 bit real mode data segment */
	.word	0xffff, 0x0000
	.byte	0x00, 0x93, 0x00, 0x00
gdt_end:
lm_exit_addr:
reloc4:	       
	.long lm_exit - _reloc
	.long 0x10
	
#if 0
#include "x86-setup-legacy-pic.S"
#include "x86-setup-legacy-timer.S"
#endif
setup16_end:
setup16_size:
	.long setup16_end - setup16_start
setup16_align:
	.long 16
